# postponed evaluation of annotation
from __future__ import annotations
import logging

logger = logging.getLogger(__name__)


VALID_SOURCE = ['auth', 'id', 'userinfo']

class RBACChecker:
    is_set = True

    def __init__(self, roles_required :  list[str] = [], claim = 'roles', source = 'userinfo'):
        if roles_required == []:
            self.is_set = False
            return

        if source not in VALID_SOURCE:
            raise ValueError(f"Source must be one of {VALID_SOURCE}")
        self.source = source
        self.claim = claim
        self.roles_requires = roles_required
        

    def check_rbac(self, auth_token = {}, id_token = {}, userinfo_token = {}) -> bool:
        """Check for role claims against provided tokens.

        Parameters
        ----------
        auth_token : dict, optional
            Authorization token, by default {}
        id_token : dict, optional
            ID token, by default {}
        userinfo_token : dict, optional
            Userinfo token, by default {}
        """     

        if self.is_set == False:
            logger.debug("No role check performed.")
            return True   

        claims = {}
        # Extract appropriate claim source.
        if self.source == VALID_SOURCE[0]: #auth
            claims = auth_token
        elif self.source == VALID_SOURCE[1]: #id
            claims = id_token
        elif self.source == VALID_SOURCE[2]: #userinfo
            claims = userinfo_token

        if claims == {}:
            raise ValueError(f"No claim data passed for source {self.source}")
        logger.debug(f"Checking claim in source {self.source}")
        
        if "." in self.claim:
            source_claim_ls = self.claim.split(".")
        else:
            source_claim_ls = [self.claim]
        try:
            claim = claims[source_claim_ls.pop(0)]
            while source_claim_ls != []:
                claim = claim[source_claim_ls.pop(0)]
            logger.debug(f"Roles assigned to user {claim}")
        except KeyError:
            logger.debug("Claim not found. Not authorizing user.")
            return False

        if not isinstance(claim, list):
            claim = [claim]
        
        if(set(self.roles_requires).issubset(set(claim))):
            return True
        return False



import collections
import logging

logger = logging.getLogger(__name__)

AuthenticationResult = collections.namedtuple('AuthenticationResult',
                                              ['access_token', 'id_token_claims', 'id_token_jwt', 'userinfo_claims'])


class AuthResponseProcessError(ValueError):
    pass


class AuthResponseUnexpectedStateError(AuthResponseProcessError):
    pass


class AuthResponseUnexpectedNonceError(AuthResponseProcessError):
    pass


class AuthResponseMismatchingSubjectError(AuthResponseProcessError):
    pass


class AuthResponseErrorResponseError(AuthResponseProcessError):
    def __init__(self, error_response):
        """
        Args:
            error_response (Mapping[str, str]): OAuth error response containing 'error' and 'error_description'
        """
        self.error_response = error_response


class AuthResponseHandler:
    def __init__(self, client):
        """
        Args:
            client (flask_pyoidc.pyoidc_facade.PyoidcFacade): Client proxy to make requests to the provider
        """
        self._client = client

    def process_auth_response(self, auth_response, expected_state, expected_nonce=None):
        """
        Args:
            auth_response (Union[AuthorizationResponse, AuthorizationErrorResponse]): parsed OIDC auth response
            expected_state (str): state value included in the OIDC auth request
            expected_nonce (str): nonce value included in the OIDC auth request
        Returns:
            AuthenticationResult: All relevant data associated with the authenticated user
        """
        if 'error' in auth_response:
            raise AuthResponseErrorResponseError(auth_response.to_dict())

        if auth_response['state'] != expected_state:
            raise AuthResponseUnexpectedStateError()

        # implicit/hybrid flow may return tokens in the auth response
        access_token = auth_response.get('access_token', None)
        id_token_claims = auth_response['id_token'].to_dict() if 'id_token' in auth_response else None
        id_token_jwt = auth_response.get('id_token_jwt', None) if 'id_token_jwt' in auth_response else None

        if 'code' in auth_response:
            token_resp = self._client.token_request(auth_response['code'])
            if token_resp:
                if 'error' in token_resp:
                    raise AuthResponseErrorResponseError(token_resp.to_dict())

                access_token = token_resp['access_token']

                if 'id_token' in token_resp:
                    id_token = token_resp['id_token']
                    logger.debug('received id token: %s', id_token.to_json())

                    if id_token['nonce'] != expected_nonce:
                        raise AuthResponseUnexpectedNonceError()

                    id_token_claims = id_token.to_dict()
                    id_token_jwt = token_resp.get('id_token_jwt')

        # do userinfo request
        userinfo = self._client.userinfo_request(access_token)
        userinfo_claims = None
        if userinfo:
            userinfo_claims = userinfo.to_dict()

        if id_token_claims and userinfo_claims and userinfo_claims['sub'] != id_token_claims['sub']:
            raise AuthResponseMismatchingSubjectError('The \'sub\' of userinfo does not match \'sub\' of ID Token.')

        return AuthenticationResult(access_token, id_token_claims, id_token_jwt, userinfo_claims)

    @classmethod
    def expect_fragment_encoded_response(cls, auth_request):
        if 'response_mode' in auth_request:
            return auth_request['response_mode'] == 'fragment'

        response_type = set(auth_request['response_type'].split(' '))
        is_implicit_flow = response_type == {'id_token'} or \
                           response_type == {'id_token', 'token'}
        is_hybrid_flow = response_type == {'code', 'id_token'} or \
                         response_type == {'code', 'token'} or \
                         response_type == {'code', 'id_token', 'token'}

        return is_implicit_flow or is_hybrid_flow



import base64
import json

import logging
from oic.oic import Client, RegistrationResponse, AuthorizationResponse, \
    AccessTokenResponse, TokenErrorResponse, AuthorizationErrorResponse
from oic.oic.message import ProviderConfigurationResponse
from oic.utils.authn.client import CLIENT_AUTHN_METHOD
from flask import url_for

logger = logging.getLogger(__name__)


class _ClientAuthentication:
    def __init__(self, client_id, client_secret):
        self._client_id = client_id
        self._client_secret = client_secret

    def __call__(self, method, request):
        """
        Args:
            method (str): Client Authentication Method. Only 'client_secret_basic' and 'client_secret_post' is
                supported.
            request (MutableMapping[str, str]): Token request parameters. This may be modified, i.e. if
                'client_secret_post' is used the client credentials will be added.

        Returns:
            (Mapping[str, str]): HTTP headers to be included in the token request, or `None` if no extra HTTPS headers
            are required for the token request.
        """
        if method == 'client_secret_post':
            request['client_id'] = self._client_id
            request['client_secret'] = self._client_secret
            return None  # authentication is in the request body, so no Authorization header is returned

        # default to 'client_secret_basic'
        credentials = '{}:{}'.format(self._client_id, self._client_secret)
        basic_auth = 'Basic {}'.format(base64.urlsafe_b64encode(credentials.encode('utf-8')).decode('utf-8'))
        return {'Authorization': basic_auth}


class PyoidcFacade:
    """
    Wrapper around pyoidc library, coupled with config for a simplified API for flask-pyoidc.
    """

    def __init__(self, provider_configuration, redirect_uri):
        """
        Args:
            provider_configuration (flask_pyoidc.provider_configuration.ProviderConfiguration)
        """
        self._provider_configuration = provider_configuration
        self._client = Client(client_authn_method=CLIENT_AUTHN_METHOD)

        provider_metadata = provider_configuration.ensure_provider_metadata()
        self._client.handle_provider_config(ProviderConfigurationResponse(**provider_metadata.to_dict()),
                                            provider_metadata['issuer'])

        if self._provider_configuration.registered_client_metadata:
            client_metadata = self._provider_configuration.registered_client_metadata.to_dict()
            registration_response = RegistrationResponse(**client_metadata)
            self._client.store_registration_info(registration_response)

        self._redirect_uri = redirect_uri

    def is_registered(self):
        return bool(self._provider_configuration.registered_client_metadata)

    def register(self, extra_registration_params=None):
        client_metadata = self._provider_configuration.register_client([self._redirect_uri], extra_registration_params)
        logger.debug('client registration response: %s', client_metadata)
        self._client.store_registration_info(RegistrationResponse(**client_metadata.to_dict()))

    def authentication_request(self, state, nonce, redirect_uri, extra_auth_params):
        """

        :param state:
        :param nonce:
        :param extra_auth_params:
        Returns:
            str: Authentication request as a URL to redirect the user to the provider.
        """
        args = {
            'client_id': self._client.client_id,
            'response_type': 'code',
            'scope': ['openid'],
            'redirect_uri': redirect_uri,
            'state': state,
            'nonce': nonce,
        }

        args.update(self._provider_configuration.auth_request_params)
        args.update(extra_auth_params)
        auth_request = self._client.construct_AuthorizationRequest(request_args=args)
        logger.debug('sending authentication request: %s', auth_request.to_json())

        return auth_request.request(self._client.authorization_endpoint)

    def parse_authentication_response(self, response_params):
        """
        Args:
            response_params (Mapping[str, str]): authentication response parameters
        Returns:
            Union[AuthorizationResponse, AuthorizationErrorResponse]: The parsed authorization response
        """
        auth_resp = self._parse_response(response_params, AuthorizationResponse, AuthorizationErrorResponse)
        if 'id_token' in response_params:
            auth_resp['id_token_jwt'] = response_params['id_token']
        return auth_resp

    def token_request(self, authorization_code):
        """
        Makes a token request.  If the 'token_endpoint' is not configured in the provider metadata, no request will
        be made.

        Args:
            authorization_code (str): authorization code issued to client after user authorization

        Returns:
            Union[AccessTokenResponse, TokenErrorResponse, None]: The parsed token response, or None if no token
            request was performed.
        """
        if not self._client.token_endpoint:
            return None

        redirect_uri = url_for(self._redirect_uri, _external=True)

        request = {
            'grant_type': 'authorization_code',
            'code': authorization_code,
            'redirect_uri': redirect_uri
        }

        logger.debug('making token request: %s', request)
        client_auth_method = self._client.registration_response.get('token_endpoint_auth_method', 'client_secret_basic')
        auth_header = _ClientAuthentication(self._client.client_id, self._client.client_secret)(client_auth_method,
                                                                                                request)
        resp = self._provider_configuration.requests_session \
            .post(self._client.token_endpoint,
                  data=request,
                  headers=auth_header) \
            .json()
        logger.debug('received token response: %s', json.dumps(resp))

        token_resp = self._parse_response(resp, AccessTokenResponse, TokenErrorResponse)
        if 'id_token' in resp:
            token_resp['id_token_jwt'] = resp['id_token']

        return token_resp

    def userinfo_request(self, access_token):
        """
        Args:
            access_token (str): Bearer access token to use when fetching userinfo

        Returns:
            oic.oic.message.OpenIDSchema: UserInfo Response
        """
        http_method = self._provider_configuration.userinfo_endpoint_method
        if not access_token or http_method is None or not self._client.userinfo_endpoint:
            return None

        logger.debug('making userinfo request')
        userinfo_response = self._client.do_user_info_request(method=http_method, token=access_token)
        logger.debug('received userinfo response: %s', userinfo_response.to_json())

        return userinfo_response

    @property
    def session_refresh_interval_seconds(self):
        return self._provider_configuration.session_refresh_interval_seconds

    @property
    def provider_end_session_endpoint(self):
        provider_metadata = self._provider_configuration.ensure_provider_metadata()
        return provider_metadata.get('end_session_endpoint')

    @property
    def post_logout_redirect_uris(self):
        return self._client.registration_response.get('post_logout_redirect_uris')

    def _parse_response(self, response_params, success_response_cls, error_response_cls):
        if 'error' in response_params:
            response = error_response_cls(**response_params)
        else:
            response = success_response_cls(**response_params)
            response.verify(keyjar=self._client.keyjar)
        return response



#   Copyright 2021 Kevin Matlock
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0

#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import functools
import json
import logging
from urllib.parse import parse_qsl, urlparse

import flask
import importlib_resources
from flask import abort, current_app
from flask.helpers import url_for
from oic import rndstr
from oic.oic.message import EndSessionRequest
from werkzeug.utils import redirect
import jwt

from .auth_response_handler import (AuthResponseErrorResponseError,
                                    AuthResponseHandler,
                                    AuthResponseProcessError)
from .provider_configuration import ClientMetadata, ProviderConfiguration
from .pyoidc_facade import PyoidcFacade
from .user_session import UninitialisedSession, UserSession
from .rbac_checker import RBACChecker

logger = logging.getLogger(__name__)


class NoAuthenticationError(Exception):
    pass


class OIDCAuthentication:
    """
    OIDCAuthentication object for Flask extension.
    """

    def __init__(self, provider_configurations=None, app=None):
        """
        Args:
            provider_configurations (Mapping[str, ProviderConfiguration]):
                provider configurations by name
            app (flask.app.Flask): optional Flask app
        """
        self._provider_configurations = provider_configurations
        if self._provider_configurations is not None:
            # Gets first provider configuration as default.
            self.default_provider = next(iter(self._provider_configurations))
        else:
            self.default_provider = None

        self.clients = None
        self._logout_view = None
        self._error_view = None
        self._redirect_uri_endpoint = None

        if app:
            self.init_app(app)

    def init_app(self, app):
        oidc_providers = app.config.get("OIDC_PROVIDERS", None)
        self.unauthenticated = app.config.get("UNAUTHENTICATED", False)
        required_roles = app.config.get("OIDC_REQUIRED_ROLES", [])
        if required_roles:
            if not (type(required_roles) is list):
                required_roles = [required_roles]
        role_claim = app.config.get("OIDC_ROLE_CLAIM", None)
        source = app.config.get("OIDC_ROLE_SOURCE", "auth")
        self.rbac_checker = RBACChecker(required_roles, role_claim, source)

        if oidc_providers is not None:
            if self._provider_configurations is None:
                self._provider_configurations = {}

            if isinstance(oidc_providers, str):
                oidc_providers = [oidc_providers]

            for provider in oidc_providers:
                if str(provider).lower() == "unauthenticated":
                    self.unauthenticated = True
                    continue

                provider_config = dict_to_providers_config(provider, app.config)
                self._provider_configurations[provider] = provider_config

            if self._provider_configurations == {}:
                self._provider_configurations = None

        if self._provider_configurations is not None:
            # Gets first provider configuration as default.
            self.default_provider = next(iter(self._provider_configurations))
        elif self.unauthenticated:
            logger.warning("Running server in unauthenticated mode.")
            self.default_provider = None
        else:
            raise NoAuthenticationError("No providers passed and UNAUTHENTICATED set to False.")

        if self._provider_configurations is not None:
            self._redirect_uri_endpoint = app.config.get(
                "OIDC_REDIRECT_ENDPOINT", "redirect_uri"
            ).lstrip("/")
            # Change from relative to absolute url depending on input.
            parse_result = urlparse(self._redirect_uri_endpoint)
            route = parse_result.path.lstrip("/")
            self.route = route
            logging.info("Authentication routing to: " + route)

            # setup redirect_uri as a flask route
            app.add_url_rule(
                "/" + route, route, self._handle_authentication_response, methods=["GET", "POST"],
            )

            # dynamically add the Flask redirect uri to the client info
            with app.app_context():
                redirect_uri = self._redirect_uri_endpoint
                logging.info("Redirect URI set to :" + redirect_uri)

                self.clients = {
                    name: PyoidcFacade(configuration, redirect_uri)
                    for (name, configuration) in self._provider_configurations.items()
                }

    def _get_post_logout_redirect_uri(self, client):
        if client.post_logout_redirect_uris:
            return client.post_logout_redirect_uris[0]
        return self._get_url_for_logout_view()

    def _get_url_for_logout_view(self):
        return url_for(self._logout_view.__name__, _external=True) if self._logout_view else None

    def _register_client(self, client):
        def default_post_logout_redirect_uris():
            url_for_logout_view = self._get_url_for_logout_view()
            if url_for_logout_view:
                return [url_for_logout_view]
            return []

        client_registration_args = {}
        post_logout_redirect_uris = client._provider_configuration._client_registration_info.get(
            "post_logout_redirect_uris", default_post_logout_redirect_uris()
        )
        if post_logout_redirect_uris:
            logger.debug(
                "registering with post_logout_redirect_uris=%s", post_logout_redirect_uris,
            )
            client_registration_args["post_logout_redirect_uris"] = post_logout_redirect_uris
        client.register(client_registration_args)

    def _authenticate(self, client, interactive=True):
        if not client.is_registered():
            self._register_client(client)

        flask.session["destination"] = flask.request.url
        flask.session["state"] = rndstr()
        flask.session["nonce"] = rndstr()

        # Use silent authentication for session refresh
        # This will not show login prompt to the user
        extra_auth_params = {}
        if not interactive:
            extra_auth_params["prompt"] = "none"

        redirect_uri = url_for(self.route, _external=True)

        login_url = client.authentication_request(
            flask.session["state"], flask.session["nonce"], redirect_uri, extra_auth_params,
        )

        auth_params = dict(parse_qsl(login_url.split("?")[1]))
        flask.session[
            "fragment_encoded_response"
        ] = AuthResponseHandler.expect_fragment_encoded_response(auth_params)
        return redirect(login_url)

    def _handle_authentication_response(self):
        has_error = flask.request.args.get("error", False, lambda x: bool(int(x)))
        if has_error:
            if "error" in flask.session:
                return self._show_error_response(flask.session.pop("error"))
            return "Something went wrong."

        if flask.session.pop("fragment_encoded_response", False):
            return importlib_resources.read_binary("flask_pyoidc", "parse_fragment.html").decode(
                "utf-8"
            )

        is_processing_fragment_encoded_response = flask.request.method == "POST"

        if is_processing_fragment_encoded_response:
            auth_resp = flask.request.form
        else:
            auth_resp = flask.request.args

        client = self.clients[UserSession(flask.session).current_provider]

        authn_resp = client.parse_authentication_response(auth_resp)
        logger.debug("received authentication response: %s", authn_resp.to_json())

        try:
            auth_handler = AuthResponseHandler(client)
            result = auth_handler.process_auth_response(
                authn_resp, flask.session.pop("state"), flask.session.pop("nonce")
            )
        except AuthResponseErrorResponseError as e:
            return self._handle_error_response(
                e.error_response, is_processing_fragment_encoded_response
            )
        except AuthResponseProcessError as e:
            return self._handle_error_response(
                {"error": "unexpected_error", "error_description": str(e)},
                is_processing_fragment_encoded_response,
            )

        if current_app.config.get("OIDC_SESSION_PERMANENT", True):
            flask.session.permanent = True

        UserSession(flask.session).update(
            result.access_token,
            result.id_token_claims,
            result.id_token_jwt,
            result.userinfo_claims,
        )

        destination = flask.session.pop("destination")
        if is_processing_fragment_encoded_response:
            # if the POST request was from the JS page handling fragment encoded responses we need to return
            # the destination URL as the response body
            return destination

        return redirect(destination)

    def _handle_error_response(self, error_response, should_redirect=False):
        if should_redirect:
            # if the current request was from the JS page handling fragment encoded responses we need to return
            # a URL for the error page to redirect to
            flask.session["error"] = error_response
            return "/" + self._redirect_uri_endpoint + "?error=1"
        return self._show_error_response(error_response)

    def _show_error_response(self, error_response):
        logger.error(json.dumps(error_response))
        if self._error_view:
            error = {
                k: error_response[k] for k in ["error", "error_description"] if k in error_response
            }
            return self._error_view(**error)

        return "Something went wrong with the authentication, please try to login again."

    def oidc_auth(self, provider_name: str = None, bearer=False):
        """Authentication decorater. 
        
        Wraps a given Flask endpoint with OIDC authentication for the given provider.
        
        Args:
            provider_name str:
                Name of the provider. If None, goes to default authentication.
            bearer bool:
                Whether access using a bearer token is allowed. Default is False.

        Raises:
            ValueError: 
                If invalid provider name is provided.
        """

        def oidc_decorator(view_func):
            PROVIDER_DECORATOR = provider_name
            BEARER = bearer

            @functools.wraps(view_func)
            def wrapper(*args, **kwargs):
                PROVIDER_NAME = PROVIDER_DECORATOR
                BEARER_ALLOWED = BEARER

                if PROVIDER_NAME is None:
                    PROVIDER_NAME = self.default_provider
                elif PROVIDER_NAME not in self._provider_configurations:
                    raise ValueError(
                        "Provider name '{}' not in configured providers: {}.".format(
                            provider_name, self._provider_configurations.keys()
                        )
                    )

                if PROVIDER_NAME is None:
                    return view_func(*args, **kwargs)

                session = UserSession(flask.session, PROVIDER_NAME)
                client = self.clients[session.current_provider]

                if BEARER_ALLOWED and "Authorization" in flask.request.headers.keys():
                    auth_msg = str(flask.request.headers["Authorization"])

                    if auth_msg.startswith("Bearer "):
                        logger.debug("accessing through bearer token")
                        token = auth_msg[7:]
                        userinfo = client.userinfo_request(token)
                        if "error" in userinfo:
                            return flask.Response(
                                "Invalid Bearer Token",
                                401,
                                {"WWW-Authenticate": 'Basic realm="Login Required"'},
                            )

                        session.update(access_token=token, userinfo=userinfo.to_dict())
                        flask.g.bearer = True
                        flask.g.access_token = token
                        flask.g.userinfo = userinfo
                        # Acess token is verified by calling the "userinfo" endpoint so no need to verify JWT.
                        access_token_claimset = {}
                        if session.access_token:
                            access_token_claimset = jwt.decode(token, options={"verify_signature": False})

                        valid_rbac = self.rbac_checker.check_rbac(access_token_claimset, id_token = {}, userinfo_token = userinfo)
                        if not valid_rbac:
                            abort(401)
                        return view_func(*args, **kwargs)

                if session.is_authenticated():
                    flask.g.bearer = False
                    logger.debug("user is authenticated")

                    # Acess token is verified by calling the "userinfo" endpoint so no need to verify JWT.
                    access_token_claimset = {}
                    if session.access_token:
                        access_token_claimset = jwt.decode(session.access_token, options={"verify_signature": False})

                    valid_rbac = self.rbac_checker.check_rbac(access_token_claimset, session.id_token, session.userinfo)
                    if not valid_rbac:
                        abort(401)
                    return view_func(*args, **kwargs)
                else:
                    logger.debug("user not authenticated, start flow")
                    return self._authenticate(client)

            return wrapper

        return oidc_decorator

    def _logout(self):
        logger.debug("user logout")
        try:
            session = UserSession(flask.session)
        except UninitialisedSession as e:
            logger.info("user was already logged out, doing nothing")
            return None

        id_token_jwt = session.id_token_jwt
        client = self.clients[session.current_provider]
        session.clear()

        if client.provider_end_session_endpoint:
            flask.session["end_session_state"] = rndstr()

            end_session_request = EndSessionRequest(
                id_token_hint=id_token_jwt,
                post_logout_redirect_uri=self._get_post_logout_redirect_uri(client),
                state=flask.session["end_session_state"],
            )

            logger.debug("send endsession request: %s", end_session_request.to_json())

            return redirect(end_session_request.request(client.provider_end_session_endpoint), 303)
        return None

    def oidc_logout(self, view_func):
        self._logout_view = view_func

        @functools.wraps(view_func)
        def wrapper(*args, **kwargs):
            if "state" in flask.request.args:
                # returning redirect from provider
                if flask.request.args["state"] != flask.session.pop("end_session_state"):
                    logger.error(
                        "Got unexpected state '%s' after logout redirect.",
                        flask.request.args["state"],
                    )
                return view_func(*args, **kwargs)

            redirect_to_provider = self._logout()
            if redirect_to_provider:
                return redirect_to_provider

            return view_func(*args, **kwargs)

        return wrapper

    def error_view(self, view_func):
        self._error_view = view_func
        return view_func


def dict_to_providers_config(provider_name: str, config_dict: dict) -> ProviderConfiguration:
    """Configures a provider from a simple Python Dictionary.
    
    Args:
        provider_name (str): Name of the provider. Must match the items in the config_dict.
        config_dict (dict): Dictionary to retrieve configurations from.
    
    Returns:
        ProviderConfiguration: Configured provider
        [description]
    """
    client_name = config_dict.get(provider_name + "_ISSUER")

    client_id = config_dict.get(provider_name + "_CLIENT", None)
    client_secret = config_dict.get(provider_name + "_SECRET", None)
    client_metadata = ClientMetadata(client_id=client_id, client_secret=client_secret)

    return ProviderConfiguration(issuer=client_name, client_metadata=client_metadata)

